import bcrypt
from sqlalchemy import Boolean, Column, DateTime, String
from sqlalchemy.orm import validates, relationship
from validate_email import validate_email

from mvapi.models import BaseModel, BaseQuery
from mvapi.web.libs.exceptions import AppValueError


class User(BaseModel):
    email = Column(String(128), nullable=False, unique=True, index=True)
    password = Column(String(128), nullable=False)
    deleted = Column(DateTime, index=True, default=None)
    is_admin = Column(Boolean, nullable=False, default=False)
    name = Column(String(128))

    sessions = relationship('Session', lazy='dynamic', viewonly=True,
                            query_class=BaseQuery)

    def __repr__(self):
        return f'<Email: {self.email}>'

    @property
    def salt(self):
        return self.email

    # noinspection PyUnusedLocal
    @validates('email')
    def validate_email(self, key, email):
        if not email:
            return email

        if not validate_email(email):
            raise AppValueError('Email is not valid')

        return email.lower()

    # noinspection PyUnusedLocal
    @validates('password')
    def validate_password(self, key, password):
        hashed = bcrypt.hashpw(password.encode(), bcrypt.gensalt())
        return hashed.decode()

    def passwords_matched(self, password):
        if not self.password:
            return False
        return bcrypt.checkpw(password.encode(), self.password.encode())



import importlib
import json
import time

from flask import Flask, g, request
from sqlalchemy import event
from sqlalchemy.engine.base import Engine
from werkzeug.exceptions import HTTPException

from mvapi.libs.database import db
from mvapi.libs.error import save_error
from mvapi.libs.exceptions import ModelKeyError, NotFoundError
from mvapi.libs.misc import import_object
from mvapi.settings import settings
from mvapi.web.libs.exceptions import AccessDeniedError, AppException, \
    AppValueError, BadRequestError, NoConverterException, \
    NoExtensionException, NotAllowedError, UnauthorizedError, \
    UnexpectedArgumentsError
from mvapi.web.libs.jsonwebtoken import JSONWebToken, JWTError
from mvapi.web.libs.logger import logger


class AppFactory:
    __instance = None
    app = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(AppFactory, cls).__new__(cls)
            cls.__instance.__create_app()

        return cls.__instance

    def __create_app(self):
        self.app = Flask(settings.APP_NAME)
        self.app.config.from_object(settings)

        self.__register_converters()
        self.__bind_extensions()
        self.__register_blueprints()

        if settings.EMAILS_MODULE:
            importlib.import_module(settings.EMAILS_MODULE)

    def __bind_extensions(self):
        extensions = settings.EXTENSIONS + [
            'mvapi.web.libs.extensions.cors',
        ]

        for ext_path in extensions:
            try:
                obj = import_object(ext_path)
            except ImportError:
                raise NoExtensionException(f'No {ext_path} extension found')

            if hasattr(obj, 'init_app') and callable(obj.init_app):
                obj.init_app(self.app)
            elif callable(obj):
                obj(self.app)
            else:
                raise NoExtensionException(
                    f'{ext_path} extension has no init_app.'
                )

            ext_name = ext_path.split('.')[-1]
            if ext_name not in self.app.extensions:
                self.app.extensions[ext_name] = obj

    def __register_blueprints(self):
        blueprints = settings.BLUEPRINTS + [
            'mvapi.web.urls.api_bp',
        ]

        for blueprint_path in blueprints:
            try:
                obj = import_object(blueprint_path)
                self.app.register_blueprint(obj)

            except ImportError:
                raise NoExtensionException(
                    f'No {blueprint_path} blueprint found'
                )

    def __register_converters(self):
        for name, path in settings.CONVERTERS:
            try:
                converter = import_object(path)
                self.app.url_map.converters[name] = converter

            except ImportError:
                raise NoConverterException(f'No {name} converter found')


def create_app():
    app = AppFactory().app

    if settings.DEBUG:
        @app.before_request
        def before_request():
            g.start = time.time()

        @app.after_request
        def after_request(response):
            diff = time.time() - g.start
            logger.debug(f'Request finished in {diff}')

            return response

    @app.teardown_appcontext
    def teardown_appcontext(exception):
        if exception:
            db.session.rollback()
        else:
            db.session.commit()

        db.session.remove()

    def app_error_response(exc, status, default_text):
        db.session.rollback()

        errors_text = '; '.join(exc.args) if exc.args else default_text

        if not settings.DEBUG and not (
                isinstance(exc, AppException) and
                not isinstance(exc, UnexpectedArgumentsError)):
            errors_text = default_text

        if status == 500:
            logger.error(errors_text, exc_info=True)

            if not settings.DEBUG:
                save_error()

        if request.blueprint == 'api':
            data = {
                'errors': errors_text.split('; '),
                'status': str(status)
            }

            return json.dumps(data), status, {
                'Content-Type': 'application/json; charset=utf-8'
            }
        else:
            data = save_error(False) if settings.DEBUG \
                else errors_text

            return data, status

    @app.errorhandler(Exception)
    def error_handler(exc):
        if isinstance(exc, (BadRequestError, AppValueError,
                            ModelKeyError,)):
            return app_error_response(exc, 400, 'Bad request')

        if isinstance(exc, UnauthorizedError):
            return app_error_response(exc, 401, 'Unauthorized')

        if isinstance(exc, AccessDeniedError):
            return app_error_response(exc, 403, 'Access denied')

        if isinstance(exc, (NotFoundError, UnexpectedArgumentsError,)):
            return app_error_response(exc, 404, 'Not found')

        if isinstance(exc, NotAllowedError):
            return app_error_response(exc, 405, 'Method not allowed')

        if isinstance(exc, HTTPException):
            return app_error_response(exc, exc.code, exc.name)

        return app_error_response(exc, 500, 'Unknown error')

    @app.before_request
    def get_current_user():
        g.current_user = None

        header = request.headers.get('Authorization')
        if not header:
            return None

        try:
            token_type, access_token = header.split(' ')
        except ValueError:
            raise BadRequestError('Wrong authorization header')

        if token_type.lower() != 'bearer':
            raise BadRequestError('Wrong authorization token type')

        try:
            jwt = JSONWebToken()
            g.current_user = jwt.get_user(access_token)
        except (NotFoundError, JWTError):
            return None

    if settings.DEBUG_SQL:
        # noinspection PyUnusedLocal
        @event.listens_for(Engine, 'before_cursor_execute')
        def before_cursor_execute(conn, cursor, statement, parameters, context,
                                  executemany):
            conn.info.setdefault('query_start_time', []).append(time.time())
            logger.debug(f'Start Query: {statement}. '
                         f'With parameters: {parameters}')

        # noinspection PyUnusedLocal
        @event.listens_for(Engine, 'after_cursor_execute')
        def after_cursor_execute(conn, cursor, statement, parameters, context,
                                 executemany):
            total = time.time() - conn.info['query_start_time'].pop(-1)
            logger.debug(f'Query Complete. Total Time: {str(total)}\n')

    return app



from collections import OrderedDict

from sqlalchemy.orm import Query

from mvapi.models import BaseModel
from mvapi.web.libs.misc import dict_value, url_for
from mvapi.web.models.user import User


class BaseSerializer:
    resource_type = None
    current_user: User = None
    item: BaseModel = None
    additional_relationships: set = None
    exclude_relationships: set = None
    admin_relationships: set = None
    relationships = None
    return_fields = None

    def __init__(self, relationships=None, return_fields=None):
        self.additional_relationships = self.additional_relationships or set()
        self.exclude_relationships = self.exclude_relationships or set()
        self.admin_relationships = self.admin_relationships or set()
        self.relationships = relationships

        self.return_fields = return_fields or set()
        if type(self.return_fields) is not set:
            self.return_fields = set(self.return_fields)

    def __render_relationships(self):
        if not hasattr(self.item, 'available_relationships'):
            return None

        relationships = OrderedDict([])

        keys = set(self.item.available_relationships.keys())
        keys |= self.additional_relationships
        keys -= self.exclude_relationships

        if self.admin_relationships and \
                not (self.current_user and self.current_user.is_admin):
            keys -= self.admin_relationships

        if self.return_fields:
            keys &= self.return_fields

        for key in sorted(keys, key=lambda k: k.lower()):
            relationships[key] = OrderedDict()

            rel_links = OrderedDict()

            if key not in self.additional_relationships:
                rel_links['self'] = url_for(
                    'api.api_view',
                    resource_type=self.item.plural_type,
                    resource_id=self.item.id_,
                    relationship_type=key
                )

            rel_links['related'] = url_for(
                    'api.api_view',
                    resource_type=self.item.plural_type,
                    resource_id=self.item.id_,
                    related_relationship_type=key
            )

            relationships[key]['links'] = rel_links

            attr: BaseModel = getattr(self.item, key, None)
            if attr:
                if isinstance(attr, Query):
                    if not self.relationships:
                        continue

                    path = f'{self.item.id_}.{key}'
                    items = dict_value(dict_=self.relationships, path=path)
                    if items is None:
                        continue

                    data = [self.__get_linkage(item) for item in items]

                elif isinstance(attr, list):
                    attr_: list[BaseModel] = attr
                    data = [self.__get_linkage(item) for item in attr_]

                else:
                    data = self.__get_linkage(attr)

                relationships[key]['data'] = data

        return relationships

    def __filter_fields(self, attributes):
        if not self.return_fields:
            return attributes

        filtered_attributes = OrderedDict()

        for attr, value in attributes.items():
            if attr in self.return_fields:
                filtered_attributes[attr] = value

        return filtered_attributes

    def render(self, item):
        self.item = item

        resp = OrderedDict([
            ('type', self.item.type_),
            ('id', self.item.id_)
        ])

        attributes = self.render_attributes()
        if attributes:
            resp['attributes'] = self.__filter_fields(attributes)

        relationships = self.__render_relationships()
        if relationships:
            resp['relationships'] = relationships

        links = self.render_links()
        if links:
            resp['links'] = links

        return resp

    def render_attributes(self):
        attrs = OrderedDict()

        if self.item.created_date:
            attrs['created_date'] = self.item.created_date

        return attrs

    def render_links(self):
        return OrderedDict([
            ('self', url_for('api.api_view',
                             resource_type=self.item.plural_type,
                             resource_id=self.item.id_))
        ])

    @staticmethod
    def __get_linkage(item):
        return {
            'type': item.type_,
            'id': item.id_
        }



import os
from datetime import datetime

import sys

from mvapi.libs.misc import render_template
from mvapi.settings import settings


def save_error(save_to_file=True):
    timestamp = datetime.utcnow()
    exc_type, exc_value, exc_traceback = sys.exc_info()
    python_executable = sys.executable
    python_version = sys.version_info[0:3]
    python_path = sys.path
    error_timestamp = timestamp.strftime('%b %d, %Y at %H:%M:%S')
    frames = get_traceback_frames(exc_traceback)
    last_frame = frames[-1]

    error_str = render_template('mvapi/error.tmpl', locals())

    if save_to_file:
        errors_dir = settings.ERRORS_PATH
        error_loc = last_frame['filename'].rpartition('/')[-1]
        error_ts = timestamp.strftime('%Y%m%d.%H%M%S')
        filename = os.path.join(errors_dir, f'{error_loc}.{error_ts}.html')

        if not os.path.exists(errors_dir):
            os.makedirs(errors_dir, exist_ok=True)

        with open(filename, 'wt') as error_file:
            error_file.write(error_str)

    return error_str


def get_traceback_frames(traceback):
    frames = []

    while traceback is not None:
        filename = traceback.tb_frame.f_code.co_filename
        lineno = traceback.tb_lineno

        frames.append({
            'filename': filename,
            'function': traceback.tb_frame.f_code.co_name,
            'lineno': lineno,
            'file_lines': get_file_lines(filename, lineno),
            'vars': traceback.tb_frame.f_locals.items()
        })

        traceback = traceback.tb_next

    return frames


def get_file_lines(filename, lineno):
    if not os.path.exists(filename):
        return []

    with open(filename, 'rt') as file:
        lines = file.readlines()

    if not lines:
        return []

    offset_lines = 7
    start_lineno = max(1, lineno - offset_lines) - 1
    end_lineno = lineno + offset_lines

    out_lines = []
    cur_lineno = start_lineno + 1

    for line in lines[start_lineno:end_lineno]:
        out_lines.append((cur_lineno, line))
        cur_lineno += 1

    return out_lines



from collections import OrderedDict

import itertools
from sqlalchemy.orm import Query

from mvapi.models import BaseModel
from mvapi.web.serializers import RESOURCE_SERIALIZERS


class ItemsSerializer:
    __items = None
    __relationships = None
    __current_user = None
    __return_fields = None

    def __init__(self, response, relationships=None, current_user=None):
        self.__items = response.data
        self.__relationships = relationships
        self.__current_user = current_user
        self.__return_fields = response.return_fields or {}

    def render(self):
        results = OrderedDict()

        if type(self.__items) is list:
            results['data'] = [
                self.__serialize_item(item=item,
                                      relationships=self.__relationships)
                for item in self.__items
            ]

        else:
            results['data'] = (
                None if not self.__items
                else self.__serialize_item(
                    item=self.__items, relationships=self.__relationships
                )
            )

        included = self.__generate_included(items=self.__items,
                                            relationships=self.__relationships)
        if included:
            results['included'] = included

        return results

    def __serialize_item(self, item, relationships=None):
        serializer = RESOURCE_SERIALIZERS.get(item.type_)
        if not serializer:
            return None

        ser_obj = serializer(
            relationships=relationships,
            return_fields=self.__return_fields.get(item.type_)
        )

        ser_obj.current_user = self.__current_user
        return ser_obj.render(item=item)

    def __generate_included(self, items, relationships):
        if not (items or relationships):
            return []

        if type(items) is not list:
            items = [items]

        data_ids = {item.id_ for item in items}
        included = {}

        for item in items:
            if issubclass(item.__class__, BaseModel):
                self.__include_items(item=item, included=included,
                                     data_ids=data_ids)

        relationships = relationships or {}

        for rel in relationships.values():
            for item in itertools.chain.from_iterable(rel.values()):
                if item.id_ not in included:
                    included[item.id_] = self.__serialize_item(item=item)
                    self.__include_items(item=item, included=included,
                                         data_ids=data_ids)

        return list(included.values())

    def __include_items(self, item, included, data_ids):
        for attr_name in set(item.available_relationships.keys()):
            if self.__return_fields and \
                    (item.type_ not in self.__return_fields or
                     attr_name not in self.__return_fields[item.type_]):
                continue

            attr: BaseModel = getattr(item, attr_name, None)
            if (not attr or
                    isinstance(attr, Query) or
                    attr.id_ in included or
                    attr.id_ in data_ids):
                continue

            serialized_item = self.__serialize_item(item=attr)
            if serialized_item:
                included[attr.id_] = serialized_item
            self.__include_items(item=attr, included=included,
                                 data_ids=data_ids)



import logging
import logging.config

from mvapi.settings import settings

logger = logging.getLogger(settings.ROOT_LOGGER_NAME)


def init_logger():
    if settings.LOGGING:
        logging.config.dictConfig(settings.LOGGING)



import click
import importlib
import sys
import os


@click.command('run-temp-script')
@click.argument('path')
def run_temp_script(path):
    """Run a script from the temp category"""

    sys.path.append(os.getcwd())
    mod = importlib.import_module(f'temp.{path}')
    func = getattr(mod, path.split('.')[-1])
    func()



import importlib
import os
import re
from datetime import datetime
from uuid import uuid4

import inflect
import shortuuid
from sqlalchemy import and_, Column, DateTime, inspect, String
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import ColumnProperty, Query, RelationshipProperty

import mvapi.web.models
from mvapi.libs.database import db
from mvapi.libs.exceptions import ModelKeyError, NotFoundError
from mvapi.libs.misc import classproperty
from mvapi.settings import settings


class BaseQuery(Query):
    def one(self, *filters):
        try:
            if filters:
                return self.filter(*filters).one()
            else:
                return super(BaseQuery, self).one()
        except NoResultFound:
            raise NotFoundError

    def get(self, ident):
        return self.get_by(id_=ident)

    def get_by(self, **kwargs):
        return self.filter_by(**kwargs).one()

    def apply_args(self, limit=None, offset=None, sort=None, filters=None):
        query = self

        if filters:
            query = query.filter(and_(*filters))

        if sort:
            query = query.order_by(*sort)

        if limit:
            query = query.limit(limit)
        if offset:
            query = query.offset(offset)

        return query


class BaseModel(declarative_base()):
    __abstract__ = True

    plural = None
    name_prefix = None
    default_sort = None
    query = db.session.query_property(query_cls=BaseQuery)

    id_: Column = Column('id', String(36), primary_key=True,
                         default=lambda: str(uuid4()))
    created_date: Column = Column(DateTime, nullable=False,
                                  default=datetime.utcnow)
    modified_date: Column = Column(DateTime, nullable=False,
                                   default=datetime.utcnow,
                                   onupdate=datetime.utcnow)

    @declared_attr
    def __tablename__(self):
        name = re.sub(r'(?<!^)(?=[A-Z])', '_', self.__name__).lower()
        if self.name_prefix:
            name = self.name_prefix + '_' + name
        return name

    @property
    def type_(self):
        return self.__table__.name.lower()

    @property
    def plural_type(self):
        if not self.plural:
            text = ' '.join(self.type_.split('_'))
            plural = inflect.engine().plural(text)
            self.plural = '_'.join(plural.split(' ')).lower()

        return self.plural.lower() if self.plural else self.type_

    @property
    def short_id(self):
        return shortuuid.encode(self.id_)

    @classproperty
    def available_columns(self):
        return {key for key, value in inspect(self).mapper.attrs.items()
                if isinstance(value, ColumnProperty)}

    @classproperty
    def available_relationships(self):
        attrs = inspect(self).mapper.attrs
        return {key: value.local_columns for key, value in attrs.items()
                if isinstance(value, RelationshipProperty)}

    @classproperty
    def relationship_keys(self):
        rels = self.available_relationships
        keys = set(rels.keys())

        for fields in rels.values():
            keys |= {field.name for field in fields
                     if field.name in self.available_columns}

        return keys

    @classproperty
    def required_keys(self):
        keys = set()

        for key in self.available_columns:
            attr = getattr(self, key)
            if hasattr(attr, 'nullable') and not attr.nullable:
                keys.add(key)

        for key, value in self.available_relationships.items():
            for column in value:
                if column.name in keys:
                    keys.add(key)
                    keys.remove(column.name)

        return keys

    def __init__(self, **kwargs):
        invalid_attrs = []

        for k in kwargs:
            if not hasattr(self.__class__, k):
                invalid_attrs.append(k)

        if invalid_attrs:
            if len(invalid_attrs) == 1:
                raise ModelKeyError(
                    f'Attribute {invalid_attrs[0]} doesn\'t exist'
                )
            else:
                attrs = ', '.join(invalid_attrs)
                raise ModelKeyError(f'Attributes {attrs} don\'t exist')

        missing_keys = self.required_keys - set(kwargs.keys()) - \
                       {'id_', 'created_date', 'modified_date'}
        if missing_keys:
            keys = ', '.join(missing_keys)
            raise ModelKeyError(f'Attributes {keys} can\'t be null or empty')

        super(BaseModel, self).__init__(**kwargs)

    def __setattr__(self, key, value):
        if key in self.required_keys and (value is None or value == ''):
            raise ModelKeyError(f'Attribute {key} can\'t be null or empty')

        super(BaseModel, self).__setattr__(key, value)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.id_}>'

    @classmethod
    def create(cls, **kwargs):
        obj = cls(**kwargs)

        db.session.add(obj)
        db.session.flush()

        return obj

    def delete(self):
        db.session.delete(self)
        db.session.flush()

    @classmethod
    def get_sort_fields(cls, sort: str = None):
        if not sort:
            if not cls.default_sort:
                return cls.created_date.desc()
            return cls.default_sort

        sort_items = []
        sort_cols = set()

        for item in sort.split(','):
            asc = True
            if item.startswith('-'):
                asc = False
                item = item[1:]

            nulls_first = True
            parts = item.split(':')
            if len(parts) == 2:
                if parts[1].lower() == 'last':
                    nulls_first = False
                item = parts[0]

            if item.lower() == 'id':
                item = 'id_'

            func = getattr(cls, f'get_{item}_sort_column', None)
            if not func:
                sort_cols.add(item)

            sort_items.append({
                'asc': asc,
                'nulls_first': nulls_first,
                'column': func or item,
            })

        if sort_cols & cls.available_columns != sort_cols:
            raise ModelKeyError

        order_fields = []
        for item in sort_items:
            column = item['column']
            if callable(column):
                columns = column()
            else:
                columns = [getattr(cls, item['column'])]

            for column in columns:
                column = column.asc() if item['asc'] else column.desc()
                column = (column.nullsfirst() if item['nulls_first']
                          else column.nullslast())
                order_fields.append(column)

        return order_fields


def import_models():
    models = [__package__] + \
             [mvapi.web.models.__package__] + \
             settings.MODELS

    for model_str in models:
        model = importlib.import_module(model_str)
        for file in os.listdir(os.path.dirname(model.__file__)):
            if not file.startswith('__') and file.endswith('.py'):
                name = file.rpartition('.')[0]
                importlib.import_module(f'{model.__package__}.{name}')



from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.orm import relationship

from mvapi.models import BaseModel


class Session(BaseModel):
    remote_ip: Column = Column(String(64))
    user_agent: Column = Column(String)
    user_id: Column = Column(String, ForeignKey('user.id', ondelete='CASCADE'),
                             index=True, nullable=False)

    user = relationship('User', lazy='joined', uselist=False)



import re
from copy import deepcopy

from flask import g, request
from sqlalchemy.orm import Query
from werkzeug.exceptions import BadRequest

from mvapi.libs.exceptions import NotFoundError
from mvapi.settings import settings
from mvapi.web.libs.exceptions import AccessDeniedError, BadRequestError, \
    UnauthorizedError
from mvapi.web.libs.misc import ApiResponse
from mvapi.web.models.user import User


class BaseView:
    resource_type = None
    resource_model = None

    current_user: User = None
    resource_id = None
    relationship_type = None
    related_relationship_type = None
    resource = None
    request_attrs = None
    request_relationships = None
    limit = 0
    current_page = 1
    next_page = None
    cursor = None
    common_args = None
    return_fields = None

    def __init__(self, **kwargs):
        self.current_user = g.current_user

        self.resource_id = kwargs.get('resource_id')
        self.relationship_type = kwargs.get('relationship_type')
        self.related_relationship_type = kwargs.get('related_relationship_type')

        self.limit = settings.LIMIT

    def __get_offset(self):
        return self.limit * (self.current_page - 1) if self.limit else 0

    def __process_request_args(self):
        self.common_args = {
            'limit': self.limit,
            'offset': self.__get_offset()
        }

        for key, value in request.args.items():
            key = key.lower()

            if key in ('include',):
                continue

            match = re.match(r'page\[([a-zA-Z0-9]+)]', key)
            if match:
                self.process_page_arg(page=match.group(1), value=value)
                continue

            match = re.match(r'filter\[([a-zA-Z0-9_]+)]', key)
            if match:
                self.process_filter_arg(filter_=match.group(1), value=value)
                continue

            match = re.match(r'fields\[([a-zA-Z0-9_]+)]', key)
            if match:
                self.process_fields_arg(field=match.group(1), value=value)
                continue

            method = getattr(self, f'process_{key}_arg', None)
            if method:
                method(value)
                continue

            raise BadRequestError

    def __process_json_data(self):
        self.request_attrs = {}
        self.request_relationships = {}

        try:
            json_data = request.json or {}
            json_data = json_data.get('data', {})
        except BadRequest:
            return

        if self.relationship_type:
            self.__process_relationship(relationship=self.relationship_type,
                                        data=json_data)

        else:
            for key, value in json_data.get('attributes', {}).items():
                func = getattr(self, f'get_{key}_attribute', None)
                self.request_attrs[key] = func(value) if func else value

            for key, value in json_data.get('relationships', {}).items():
                self.__process_relationship(relationship=key,
                                            data=value['data'])

        if self.resource_model:
            self.__check_nullables()

    def __process_relationship(self, relationship, data):
        if self.resource_model:
            rel_attr = getattr(self.resource_model, relationship, None)
            if (not rel_attr or
                    (type(data) is list) != rel_attr.property.uselist):
                raise BadRequestError

        func = getattr(self, f'get_{relationship}_relationship', None)
        if not func:
            raise NotImplementedError

        if not data:
            results = data

        elif type(data) is list:
            try:
                results = func([item['id'] for item in data])
            except NotFoundError:
                raise BadRequestError

            if len(results) != len(data):
                raise BadRequestError

            if type(results) is not list:
                results = [results]

        else:
            try:
                results = func(data['id'])
            except NotFoundError:
                raise BadRequestError

            if not results:
                raise BadRequestError

            if type(results) is list:
                results = results[0]

        self.request_relationships[relationship] = results

    def __check_nullables(self):
        for key in (self.resource_model.required_keys &
                    set(self.request_attrs.keys())):
            value = self.request_attrs[key]

            if type(value) is bool and value is False:
                continue

            if not value:
                raise BadRequestError(
                    f'Attribute {key} can\'t be null or empty'
                )

        for key in (self.resource_model.required_keys &
                    set(self.request_relationships.keys())):
            if not self.request_relationships[key]:
                raise BadRequestError(
                    f'Relationship {key} can\'t be null or empty'
                )

    def process_request(self):
        req_method = request.method.lower()
        self.__process_request_args()

        if (req_method == 'post' and self.resource_id and
                not self.relationship_type and
                not self.related_relationship_type):
            raise NotFoundError

        if self.related_relationship_type:
            if req_method not in ('head', 'get'):
                raise NotFoundError

            self.resource = self.get_resource()
            if not self.resource:
                raise NotFoundError

            f_name = f'process_{self.related_relationship_type}_relationship'
            func = getattr(self, f_name, None)
            if func:
                return func()

            attr = getattr(self.resource, self.related_relationship_type, None)
            if attr:
                if isinstance(attr, Query):
                    return attr.apply_args(**self.common_args)
                else:
                    return attr
            else:
                raise NotFoundError

        method_name = req_method
        if self.relationship_type:
            if req_method not in ('post', 'patch', 'delete'):
                raise NotFoundError

            method_name = f'{req_method}_{self.relationship_type}_relationship'

        method = getattr(self, method_name, None)
        if method is None and req_method == 'head':
            method = getattr(self, 'get', None)

        if not method:
            raise NotFoundError

        if (req_method in ('post', 'patch', 'delete',) and
                not request.files):
            self.__process_json_data()

        # Disallow updating a whole relationship collection for a resource
        # when patching the resource itself
        if not self.relationship_type and req_method in ('patch',):
            for value in self.request_relationships.values():
                if type(value) is list:
                    raise BadRequestError

        if self.resource_id:
            self.resource = self.get_resource()

        data = method()

        meta = None
        if type(data) is tuple:
            data, meta = data

        if data is None:
            data = self.resource

        if isinstance(data, Query):
            data = data.all()

        return ApiResponse(
            data=data,
            meta=meta,
            limit=self.limit,
            current_page=self.current_page,
            next_page=self.next_page,
            cursor=self.cursor,
            relationship_type=self.relationship_type,
            related_relationship_type=self.related_relationship_type,
            return_fields=self.return_fields
        )

    def available_data(self, include=None, exclude=None, required=None,
                       extra=None):
        exclude_columns = {'id', 'created_date', 'modified_date'}
        exclude_columns -= include or set()
        exclude_columns |= exclude or set()

        attr_keys, rel_keys = set(), set()
        if self.resource_model:
            attr_keys = self.resource_model.available_columns
            rel_keys = self.resource_model.relationship_keys

        attr_keys |= include or set()
        attr_keys |= required or set()
        attr_keys |= extra or set()

        data = {
            k: deepcopy(v) for k, v in self.request_attrs.items()
            if not attr_keys or (k in attr_keys and k not in exclude_columns)
        }

        data.update({
            k: deepcopy(v)
            for k, v in self.request_relationships.items()
            if not rel_keys or (k in rel_keys and type(v) is not list)
        })

        if required:
            errors = []
            for key in required:
                val = data.get(key)
                if not val and key in self.resource_model.available_columns:
                    column_name = self.resource_model.readable_column_name(key)
                    errors.append(f'{column_name} is required')

            if errors:
                raise BadRequestError('; '.join(errors))

        for attr, value in data.items():
            func = getattr(self, f'check_{attr}_value', None)
            if func:
                func(value)

        return data

    def drop_relationships(self, name, existing=False, missing=False,
                           resource=None):
        resource = resource or self.resource
        rel_items = self.request_relationships[name]
        relationship = getattr(resource, name)
        relationship_cls = relationship.attr.target_mapper.class_

        items = relationship.apply_args(limit=0, filters=[
            relationship_cls.id_.in_({i.id_ for i in rel_items})
        ])

        item_ids = {i.id_ for i in items}

        if existing:
            return filter(lambda i: i.id_ not in item_ids, rel_items)
        if missing:
            return filter(lambda i: i.id_ in item_ids, rel_items)

        return rel_items

    def check_owner(self, resource_user: User):
        if not self.current_user:
            raise UnauthorizedError

        if self.current_user.id_ != resource_user.id_ and \
                not self.current_user.is_admin:
            raise AccessDeniedError

    def get_resource_owner(self, param_name=None):
        return self.resource if not param_name else \
            getattr(self.resource, param_name)

    def get_resource(self):
        if self.resource:
            return self.resource

        if self.resource_model and self.resource_id:
            return self.resource_model.query.get(self.resource_id)

        raise NotFoundError

    def get_resources(self, allow_all=False):
        if allow_all or (self.current_user and self.current_user.is_admin):
            return (
                self.get_resource() if self.resource_id
                else self.resource_model.query.apply_args(**self.common_args)
            )

        else:
            if self.resource_id:
                return self.get_resource()

            raise AccessDeniedError

    def process_sort_arg(self, value):
        self.common_args['sort'] = self.resource_model.get_sort_fields(value)

    def process_page_arg(self, page, value):
        page = page.lower()

        if page == 'size':
            self.limit = int(value)
            self.common_args['limit'] = self.limit

        if page == 'number':
            self.current_page = int(value)

        self.common_args['offset'] = self.__get_offset()

    def process_filter_arg(self, filter_, value):
        filter_ = filter_.lower()
        method = getattr(self, f'get_{filter_}_filter', None)
        if not method:
            raise BadRequestError(f'Filter {filter_} is not supported')

        filters = method(value)
        if filters:
            if 'filters' not in self.common_args:
                self.common_args['filters'] = []

            self.common_args['filters'] += filters

    def process_fields_arg(self, field, value):
        if self.return_fields is None:
            self.return_fields = {}

        self.return_fields[field] = value.split(',')

    def get_ids_filter(self, ids):
        if not ids:
            return []

        ids = ids.replace(', ', ',').lower().split(',')
        return [self.resource_model.id_.in_(ids)]



from functools import wraps

from flask import g

from mvapi.libs.exceptions import NotFoundError
from mvapi.settings import settings
from mvapi.web.libs.exceptions import AccessDeniedError, UnauthorizedError


def auth_required(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if not g.current_user:
            raise UnauthorizedError

        return func(*args, **kwargs)

    return decorated_view


def admin_required(func):
    @wraps(func)
    @auth_required
    def decorated_view(*args, **kwargs):
        if not g.current_user.is_admin:
            raise AccessDeniedError

        return func(*args, **kwargs)

    return decorated_view


def debug_only(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if not settings.DEBUG:
            raise NotFoundError

        return func(*args, **kwargs)

    return decorated_view



import re

from setuptools import find_packages, setup

with open('./mvapi/version.py') as file:
    version = re.search(r'version = \'(.*?)\'', file.read()).group(1)

setup(
    name='mvapi',
    version=version,
    author='Mikhail Vetoshkin',
    author_email='mikhail@vetoshkin.dev',
    description='Skeleton for a JSON API project',
    url='https://git.vetoshkin.dev/mvapi',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'mvapi = mvapi:main'
        ]
    },
    install_requires=[
        'alembic == 1.7.5',
        'bcrypt == 3.2.0',
        'click == 8.0.3',
        'flask == 2.0.2',
        'flask-cors == 3.0.10',
        'inflect == 5.4.0',
        'jinja2 == 3.0.3',
        'psycopg2 == 2.9.3',
        'pyjwt == 2.3.0',
        'python-dateutil==2.8.2',
        'python-dotenv == 0.19.1',
        'shortuuid == 1.0.8',
        'sqlalchemy == 1.4.28',
        'validate-email == 1.3',
        'werkzeug == 2.1.0',
    ],
    python_requires='>=3.9',
)



import click
from alembic.command import init

from .utils import DIRECTORY, get_config


@click.command('init')
@click.option('--directory', default=DIRECTORY,
              help='path to the migrations directory')
@click.option('--template', default='generic',
              help='name of the migration environment template to use')
def init_(directory, template):
    """Initialize a new scripts directory"""

    config = get_config(directory)
    init(config, directory, template=template)



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from mvapi.settings import settings


class DB:
    __instance = None
    session = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(DB, cls).__new__(cls)

            engine = create_engine(settings.SQLALCHEMY_DATABASE_URI)
            cls.__instance.session = scoped_session(
                sessionmaker(autocommit=False, bind=engine)
            )

        return cls.__instance


db = DB()
